"use strict";

var _interopRequire = function (obj) { return obj && obj.__esModule ? obj["default"] : obj; };

var _createClass = (function () { function defineProperties(target, props) { for (var key in props) { var prop = props[key]; prop.configurable = true; if (prop.value) prop.writable = true; } Object.defineProperties(target, props); } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; })();

var _classCallCheck = function (instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } };

var assign = _interopRequire(require("object-assign"));

var EventEmitter = require("eventemitter3").EventEmitter;

var toposort = _interopRequire(require("toposort"));

var getAllPropertyNames = _interopRequire(require("getallpropertynames"));

var Actions = _interopRequire(require("./Actions"));

var Store = _interopRequire(require("./Store"));

var allActionsProperties = getAllPropertyNames(Actions.prototype);
var allStoreProperties = getAllPropertyNames(Store.prototype);
var eventEmitterProperties = Object.keys(EventEmitter.prototype);

var Dispatcher = (function () {

    /**
     * Constructor
     * @param  {Object} options
     * @param  {Object} options.actions Namespaced actions
     * @param  {Object} options.stores  Namespaced stores
     * @return {Dispatcher}
     */

    function Dispatcher(options) {
        _classCallCheck(this, Dispatcher);

        // Decorated actions
        this.actions = {};
        this.actionIds = [];
        // Decorated stores
        this.stores = {};
        // Actual stores
        this._stores = {};
        // Order in which actions get dispatched to the stores
        this.order = [];

        var actions = options.actions || {};
        var stores = options.stores || {};

        this.createActions(actions);
        this.createStores(stores);
    }

    _createClass(Dispatcher, {
        dispatch: {

            /**
             * Dispatches an action
             * @param  {String}    id   The id of the action (e.g. 'todos.create')
             * @param  {...mixed}  args Arguments that will be passed to the handlers
             * @return {void}
             */

            value: function dispatch(id) {
                for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
                    args[_key - 1] = arguments[_key];
                }

                if (this._isDispatching) {
                    throw new Error("Cannot dispatch in the middle of a dispatch.");
                }

                this._isDispatching = true;

                // Run through stores and invoke registered handlers
                var stores = this._stores;
                var _iteratorNormalCompletion = true;
                var _didIteratorError = false;
                var _iteratorError = undefined;

                try {
                    for (var _iterator = this.order[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                        var key = _step.value;

                        var handlers = stores[key]._handlers;
                        if (!handlers || !handlers[id]) continue;
                        handlers[id].apply(handlers, args);
                    }
                } catch (err) {
                    _didIteratorError = true;
                    _iteratorError = err;
                } finally {
                    try {
                        if (!_iteratorNormalCompletion && _iterator["return"]) {
                            _iterator["return"]();
                        }
                    } finally {
                        if (_didIteratorError) {
                            throw _iteratorError;
                        }
                    }
                }

                this._isDispatching = false;
            }
        },
        createActions: {

            /**
             * Create actions
             * @param  {Object} actions Namespaced actions
             * @return {void}
             */

            value: function createActions(actions) {
                var _this = this;

                // Run through namespaced actions
                for (var key in actions) {
                    var _iteratorNormalCompletion;

                    var _didIteratorError;

                    var _iteratorError;

                    var _iterator, _step;

                    (function (key) {
                        var Actions = actions[key];
                        // Make actions available at construction time
                        assign(Actions.prototype, { actions: _this.actions });
                        // Instantiate actions
                        var instance = new Actions();
                        // Create decorated actions object
                        _this.actions[key] = {};
                        // Find actual action function
                        var props = getAllPropertyNames(instance).filter(function (prop) {
                            // Ignore the base class properties
                            return allActionsProperties.indexOf(prop) < 0 &&
                            // Only regard functions
                            typeof instance[prop] == "function";
                        });
                        // Run through actual actions
                        _iteratorNormalCompletion = true;
                        _didIteratorError = false;
                        _iteratorError = undefined;

                        try {
                            for (_iterator = props[Symbol.iterator](); !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                                var prop = _step.value;

                                // Bind function to instance
                                var fn = instance[prop] = instance[prop].bind(instance);
                                // The action id is composed from the actions key and its function name
                                var id = [key, prop].join(".");
                                _this.actionIds.push(id);
                                // Listen to the action event
                                instance.addListener(prop, _this.dispatch.bind(_this, id));
                                // Add function to the decorated object
                                _this.actions[key][prop] = fn;
                            }
                        } catch (err) {
                            _didIteratorError = true;
                            _iteratorError = err;
                        } finally {
                            try {
                                if (!_iteratorNormalCompletion && _iterator["return"]) {
                                    _iterator["return"]();
                                }
                            } finally {
                                if (_didIteratorError) {
                                    throw _iteratorError;
                                }
                            }
                        }
                    })(key);
                }
            }
        },
        actionIdExists: {

            /**
             * Check if action id exists
             * @param  {String} id The action id
             * @return {Boolean}
             */

            value: function actionIdExists(id) {
                return this.actionIds.indexOf(id) > -1;
            }
        },
        createStores: {

            /**
             * Create stores
             * @param  {Object} stores Namespaced stores
             * @return {void}
             */

            value: function createStores(stores) {
                var _this = this;

                var nodes = [];
                var edges = [];

                // Create a dependency graph
                for (var key in stores) {
                    nodes.push(key);
                    var store = stores[key];
                    // If store is not an array, it has no dependencies
                    if (!Array.isArray(store)) continue;
                    var deps = store.slice(1) || [];
                    // Add edges between store and it dependencies
                    var _iteratorNormalCompletion = true;
                    var _didIteratorError = false;
                    var _iteratorError = undefined;

                    try {
                        for (var _iterator = deps[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                            var dep = _step.value;
                            edges.push([key, dep]);
                        }
                    } catch (err) {
                        _didIteratorError = true;
                        _iteratorError = err;
                    } finally {
                        try {
                            if (!_iteratorNormalCompletion && _iterator["return"]) {
                                _iterator["return"]();
                            }
                        } finally {
                            if (_didIteratorError) {
                                throw _iteratorError;
                            }
                        }
                    }
                }

                // Topological sort, store the order
                var order = this.order = toposort.array(nodes, edges).reverse();

                // Run through ordered stores
                for (var i = 0, l = order.length; i < l; i++) {
                    var _iteratorNormalCompletion2;

                    var _didIteratorError2;

                    var _iteratorError2;

                    var _iterator2, _step2;

                    (function (i, l) {
                        var key = order[i];
                        var Store = stores[key];
                        // Handle plain and array definition
                        if (Array.isArray(Store)) Store = Store[0];
                        // Make stores available at construction time
                        assign(Store.prototype, {
                            stores: _this.stores,
                            _actionIdExists: _this.actionIdExists.bind(_this)
                        });
                        // Instantiate the store
                        var instance = new Store();
                        _this._stores[key] = instance;
                        // Create a decorated stores object
                        _this.stores[key] = {};
                        // Find functions that will be added to the decorated object
                        var props = getAllPropertyNames(instance).filter(function (prop) {
                            // Only regard functions
                            return typeof instance[prop] === "function" && (prop.indexOf("get") === 0 || eventEmitterProperties.indexOf(prop) > -1 && prop !== "emit");
                        });
                        // Run through functions
                        _iteratorNormalCompletion2 = true;
                        _didIteratorError2 = false;
                        _iteratorError2 = undefined;

                        try {
                            for (_iterator2 = props[Symbol.iterator](); !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                                var prop = _step2.value;

                                // Bind function to the instance and add it to the decorated object
                                _this.stores[key][prop] = instance[prop].bind(instance);
                            }
                        } catch (err) {
                            _didIteratorError2 = true;
                            _iteratorError2 = err;
                        } finally {
                            try {
                                if (!_iteratorNormalCompletion2 && _iterator2["return"]) {
                                    _iterator2["return"]();
                                }
                            } finally {
                                if (_didIteratorError2) {
                                    throw _iteratorError2;
                                }
                            }
                        }
                    })(i, l);
                }
            }
        }
    });

    return Dispatcher;
})();

module.exports = Dispatcher;

// Functions that start with get

// Event emitter function, except emit